import boto3

from slai.exceptions import InvalidAWSProfile
from botocore.exceptions import ClientError, ProfileNotFound
from slai.modules.parameters import from_config
from importlib import import_module


def get_s3_client(*, aws_profile, region):
    import_path = from_config(
        "S3_CLIENT",
        "slai.clients.S3_CLIENT",
    )
    class_ = import_path.split(".")[-1]
    path = ".".join(import_path.split(".")[:-1])
    return getattr(import_module(path), class_)(aws_profile, region)


class S3Client:
    def __init__(self, *, aws_profile, aws_region):
        try:
            self.session = boto3.Session(
                profile_name=aws_profile, region_name=aws_region
            )
        except ProfileNotFound:
            raise InvalidAWSProfile("profile_not_found")

        self.aws_region = aws_region
        self.client = self.session.client("s3")

    def create_bucket(self, bucket_name, aws_region=None):
        try:
            self.client.create_bucket(Bucket=bucket_name)
        except ClientError:
            raise

        return True

    def upload(
        self, *, bucket_name, data, key, content_type="application/octet-stream"
    ):
        try:
            object = self.client.Object(
                bucket_name,
                key,
            )
            response = object.put(Body=data, ContentType=content_type)
            print("successfully uploaded file: {}".format(response))
        except ClientError as e:
            print(str(e))

        return response

    # def read_file(self, *, key):
    #     file_stream = None

    #     try:
    #         object = self.client.Object(
    #             S3_BUCKET_NAME,
    #             key,
    #         )
    #         file_stream = object.get()["Body"].read()

    #     except ClientError as e:
    #         print(str(e))

    #     return file_stream



import sys
import ast

from collections import namedtuple

Import = namedtuple("Import", ["module", "name", "alias"])
BUILT_INS = [
    "os",
    "sys",
    "collections",
    "inspect",
]  # TODO: figure out the built-ins dynamically

USE_LATEST = ["torch"]

PACKAGE_MAP = {"sklearn": "scikit-learn"}


def get_imports(path):
    with open(path) as fh:
        root = ast.parse(fh.read(), path)

    for node in ast.iter_child_nodes(root):
        if isinstance(node, ast.Import):
            module = []
        elif isinstance(node, ast.ImportFrom):
            module = node.module.split(".")
        else:
            continue

        for n in node.names:
            yield Import(module, n.name.split("."), n.asname)


def retrieve_version(*, base_module_name, _import, is_derived=False):
    import_line = None

    try:
        version = f"=={sys.modules[base_module_name].__version__}"
    except AttributeError:
        return None, None

    if base_module_name in USE_LATEST:
        version = ""

    if is_derived:
        base_name = base_module_name
        if len(_import.module) > 1:
            base_name = ".".join(_import.module)

        import_line = f"from {base_name} import {'.'.join(_import.name)}"
    else:
        import_line = f"import {'.'.join(_import.name)}"

    if _import.alias:
        import_line += f" as {_import.alias}"

    return version, import_line


def generate_requirements_from_runtime(*, module_name):
    requirements = {}
    import_lines = []

    imports = get_imports(sys.modules[module_name].__file__)

    for _import in imports:
        derived = False

        if len(_import.module) > 0:
            _base_module_name = _import.module[0]
            derived = True
        else:
            _base_module_name = _import.name[0]

        if _base_module_name in BUILT_INS:
            continue

        version, import_line = retrieve_version(
            base_module_name=_base_module_name, _import=_import, is_derived=derived
        )

        if requirements is not None and import_line is not None:
            dist_name = PACKAGE_MAP.get(_base_module_name)

            if dist_name is not None:
                _base_module_name = dist_name

            requirements[_base_module_name] = version
            import_lines.append(import_line)

    import_lines = set(import_lines)
    return requirements, import_lines



import json

from importlib import import_module
from pydrive.auth import GoogleAuth
from pydrive.drive import GoogleDrive
from slai.modules.parameters import from_config
from slai.clients.cli import get_cli_client
from slai.exceptions import IntegrationError


def get_google_drive_client():
    import_path = from_config(
        "GOOGLE_DRIVE_CLIENT",
        "slai.clients.gdrive.GoogleDriveClient",
    )
    class_ = import_path.split(".")[-1]
    path = ".".join(import_path.split(".")[:-1])
    return getattr(import_module(path), class_)()


class GoogleDriveClient:
    def __init__(self):
        self.cli_client = get_cli_client()

        gauth = self._authenticate()
        self.drive_client = GoogleDrive(gauth)

    def _authenticate(self):
        user = self.cli_client.get_user()

        if user.get("identity_user_google_auth") is not None:
            with open(".slai/gauth/creds.json", "w") as f_out:
                json.dump(user["identity_user_google_auth"]["gauth_creds"], f_out)
        else:
            raise IntegrationError("google_drive_integration_not_found")

        # TODO: handle failed auth
        gauth = GoogleAuth()
        gauth.LoadCredentialsFile(".slai/gauth/creds.json")
        gauth.Authorize()

        return gauth

    def _format_parents(self, parent_ids):
        parents = []
        if parent_ids is not None:
            for _id in parent_ids:
                parents.append({"id": _id})
        return parents

    def upload_file(self, *, filename, local_path, parent_ids, file_id=None):
        parents = self._format_parents(parent_ids)

        _file = None
        if file_id is None:
            _file = self.drive_client.CreateFile(
                {"title": f"{filename}", "parents": parents}
            )
        else:
            _file = self.drive_client.CreateFile(
                {"title": f"{filename}", "parents": parents, "id": file_id}
            )

        _file.SetContentFile(local_path)
        _file.Upload()

        return _file["id"]

    def create_folder(self, *, name, parent_ids=None):
        parents = self._format_parents(parent_ids)
        folder_file = self.drive_client.CreateFile(
            {
                "title": name,
                "parents": parents,
                "mimeType": "application/vnd.google-apps.folder",
            }
        )
        folder_file.Upload()
        return folder_file["id"]

    def download_file(self, *, file_id, local_path):
        _file = self.drive_client.CreateFile({"id": file_id})
        _file.GetContentFile(local_path)
        return _file



import os
import xxhash
import shutil
import requests
import pathlib
import errno

from http import HTTPStatus
from tqdm import tqdm
from collections import namedtuple
from slai.clients.cli import get_cli_client
from slai.exceptions import DatasetNotFound

DataSetNode = namedtuple("DataSetNode", "name parent_name hash index")


class DataSet:
    def __init__(self, name, local_path=None, public=False, version=None, create=False):
        self.name = name
        self.local_path = local_path
        self.public = public
        self.create = create
        self.version = version
        self.metatable = {}

        self.cli_client = get_cli_client()

    def add_file(self, file_path, dir=None):
        if not self.local_path:
            raise RuntimeError("download_dataset_first")

        if dir is not None:
            shutil.copy(file_path, f"{self.local_path}/{dir}/")
        else:
            shutil.copy(file_path, f"{self.local_path}/")

    def add_files(self, files=[], dir=None):
        for f in files:
            self.add_file(f, dir)

    def add_directory(self, *, path, dir=None):
        if dir is not None:
            shutil.copytree(path, f"{self.local_path}/{dir}/{path}")
        else:
            shutil.copytree(path, f"{self.local_path}/{path}")

    def _handle_bad_request(self, response_body):
        msg = response_body["detail"]["non_field_errors"][0]
        if msg == "dataset_not_found":
            raise DatasetNotFound

    def download(self):
        try:
            self.dataset_object = self.cli_client.retrieve_dataset(
                name=self.name,
                version=self.version,
                public=self.public,
                create=self.create,
            )
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == HTTPStatus.BAD_REQUEST:
                self._handle_bad_request(e.response.json())
                return
            else:
                raise

        self.metatable = self._load_metatable()
        self.diff_table = {}

        if not self.local_path:
            self.local_path = f"./{self.name}"
            pathlib.Path(self.local_path).mkdir(parents=True, exist_ok=True)

        for file_id, obj in tqdm(self.dataset_object["object_urls"].items()):
            self._fetch_file(url=obj["url"], path=obj["path"])

    def _fetch_file(self, *, url, path):
        try:
            r = requests.get(url, stream=True)
        except requests.exceptions.ConnectionError:
            return False

        if r.status_code == HTTPStatus.OK:
            file = f"{self.local_path}/{path}"

            if not os.path.exists(os.path.dirname(file)):
                try:
                    os.makedirs(os.path.dirname(file))
                except OSError as exc:
                    if exc.errno != errno.EEXIST:
                        raise

            with open(f"{file}", "wb") as f:
                for chunk in r:
                    f.write(chunk)

        return True

    def _put_file(self, *, url, path):
        with open(f"{self.local_path}/{path}", "rb") as f:
            try:
                r = requests.put(
                    f"{url}",
                    data=f.read(),
                )
            except requests.exceptions.ConnectionError:
                return False

            r.raise_for_status()

        return True

    def path(self):
        return self.local_path

    def diff(self):
        self._checksum_local_path()

        for file_id, value in self.diff_table.items():
            print(f"{file_id}: {value}")

    def save(self):
        changes = self._checksum_local_path()
        if changes:
            self.dataset_object = self.cli_client.update_dataset(
                name=self.name, public=self.public, diff_table=self.diff_table
            )

            for file_id, obj in tqdm(self.dataset_object["object_urls"].items()):
                self._put_file(url=obj["url"], path=obj["path"])

            self.metatable = self._load_metatable()
            self.diff_table = {}

    def describe(self):
        if not self.metatable:
            return

        return self.metatable.values()

    def reset(self):
        raise NotImplementedError

    def _load_metatable(self):
        metatable = {}
        for file_id, node in self.dataset_object["metatable"].items():
            node = DataSetNode(
                name=node[0], parent_name=node[1], hash=node[2], index=node[3]
            )
            metatable[file_id] = node

        return metatable

    def _regenerate_metatable(self):
        metatable = {}
        for dirpath, _, fnames in os.walk(self.local_path):
            path = dirpath.replace(str(self.local_path), "", 1)
            if len(path) > 0 and path[0] == "/":
                path = path[1:]

            for f in fnames:
                node = DataSetNode(
                    name=f,
                    parent_name=path,
                    hash=xxhash.xxh128(
                        open(f"{self.local_path}/{path}/{f}", "rb").read()
                    ).hexdigest(),
                    index=0,
                )
                file_id = xxhash.xxh128_hexdigest(f"{path}/{f}")
                metatable[file_id] = node

        return metatable

    def _checksum_local_path(self):
        self.diff_table = {}

        changes = False

        if not self.local_path:
            raise RuntimeError("download_dataset_first")

        self.next_metatable = self._regenerate_metatable()

        for file_id, node in self.next_metatable.items():
            if file_id in self.metatable:
                existing_node = self.metatable[file_id]

                if node.hash != existing_node.hash:
                    node = node._replace(index=existing_node.index + 1)
                    self.diff_table[file_id] = {
                        "node": node._asdict(),
                        "action": "modified",
                    }
                    changes = True
            else:
                node = node._replace(index=0)

                self.diff_table[file_id] = {
                    "node": node._asdict(),
                    "action": "created",
                }
                changes = True

        file_ids = set(self.metatable.keys())
        next_file_ids = set(self.next_metatable.keys())

        deleted_file_ids = file_ids.difference(next_file_ids)

        if len(deleted_file_ids) > 0:
            changes = True

        for file_id in list(deleted_file_ids):
            self.diff_table[file_id] = {
                "node": self.metatable[file_id]._asdict(),
                "action": "deleted",
            }

        return changes



import json
import tokenize
import re
import datetime

from pathlib import Path
from jinja2 import Template
from slai.constants import NOTEBOOK_TRAINER_PREFIX, DEFAULT_TRAINER_TIMEOUT_S


def parse_notebook(notebook_path):
    """
    Parses notebook contents into a model trainer.
    """

    notebook_data = None

    with open(notebook_path, "r") as f_in:
        notebook_data = json.load(f_in)

    output_source = []

    # only use cells that contain the NOTEBOOK_TRAINER_PREFIX comment
    for cell in notebook_data["cells"]:
        cell_source_line_iter = iter(cell["source"])

        tokens = list(tokenize.generate_tokens(lambda: next(cell_source_line_iter)))

        for token_type, token_value, begin, end, line in tokens:
            if token_type == tokenize.COMMENT:
                m = re.match(NOTEBOOK_TRAINER_PREFIX, line)
                if m:
                    output_source.extend(cell["source"])
                    output_source[-1] = output_source[-1].strip() + "\n"
                    break

    import_lines = list(
        filter(
            lambda l: l.startswith("import ") or l.startswith("from "), output_source
        )
    )

    for idx, _ in enumerate(import_lines):
        import_lines[idx] = import_lines[idx].strip() + "\n"

    output_source = list(
        filter(
            lambda l: not l.startswith("import ") and not l.startswith("from "),
            output_source,
        )
    )

    output_source = list(
        filter(lambda l: NOTEBOOK_TRAINER_PREFIX not in l, output_source)
    )

    for idx in range(len(output_source)):
        if output_source[idx] != "\n":
            output_source[idx] = (" " * 8) + output_source[idx]

    imports = "".join(import_lines)
    trainer = "".join(output_source)

    return imports, trainer


def generate_trainer_from_template(*, model_name, imports, trainer, model_version_id):
    slai_module_path = Path(__file__).parent.parent
    trainer_template_path = f"{slai_module_path}/templates/trainer.py.template"

    template_variables = {
        "SLAI_MODEL_NAME": model_name,
        "SLAI_TRAINER_CREATED_AT": datetime.datetime.now().isoformat(),
        "SLAI_TRAINER_IMPORTS": imports,
        "SLAI_TRAINER_CONTENTS": trainer,
        "SLAI_CURRENT_MODEL_VERSION": model_version_id,
        "SLAI_TRAINER_TIMEOUT": DEFAULT_TRAINER_TIMEOUT_S,
    }

    template_contents = None
    rendered_template = None

    with open(trainer_template_path, "r") as f_in:
        template_contents = f_in.read()
        t = Template(template_contents)
        rendered_template = t.render(**template_variables)

    return rendered_template



import pickle

from slai.loaders import base_loader


class SklearnLoader(base_loader.BaseLoader):
    @classmethod
    def load_model(cls, model_artifact, deployment_instance_path):
        model_artifact_model_binary = model_artifact

        # write model to deployment instance environment
        model_path = f"{deployment_instance_path}/model"
        with open(model_path, "wb") as f_out:
            f_out.write(model_artifact_model_binary)

        loaded_model = pickle.loads(model_artifact_model_binary)

        inference_method_name = "predict"
        return loaded_model, inference_method_name



from abc import ABC, abstractmethod
from importlib import import_module


class InvalidInputException(Exception):
    def __init__(self, msg, errors=None):
        super().__init__(msg)
        self.msg = msg
        self.errors = errors


class ModelInputSerializer(ABC):
    def __init__(self, *args, **kwargs):
        return self.define(*args, **kwargs)

    @abstractmethod
    def define(self, *args, **kwargs):
        pass

    @abstractmethod
    def serialize(self, input):
        pass


class FloatInput(ModelInputSerializer):
    def define(self):
        pass

    def serialize(self, input):
        try:
            output = float(input)
        except ValueError:
            raise InvalidInputException("invalid_float_value")

        return output


class StringInput(ModelInputSerializer):
    def define(self, max_length=None):
        self.max_length = max_length

    def serialize(self, input):
        output = str(input)

        if len(output) > self.max_length:
            raise InvalidInputException("input_string_too_long")

        return output


class NumpyInput(ModelInputSerializer):
    def define(self, shape=None, dtype=float):
        self.shape = shape
        self.dtype = dtype

    def serialize(self, input):
        try:
            np = import_module("numpy")
        except ModuleNotFoundError:
            raise RuntimeError("numpy_required_for_numpy_inputs")

        output = np.array(input, dtype=self.dtype)

        errors = []
        if self.shape is not None and output.shape != self.shape:
            errors = ["invalid_shape"]

        if errors:
            raise InvalidInputException("invalid_numpy_shape", errors=errors)

        return output


class DataframeInput(ModelInputSerializer):
    def define(self, max_rows=None, max_cols=None):
        self.max_rows = max_rows
        self.max_cols = max_cols

    def serialize(self, input):
        try:
            pd = import_module("pandas")
        except ModuleNotFoundError:
            raise RuntimeError("pandas_required_for_dataframe_inputs")

        try:
            output = pd.read_json(input)
        except ValueError:
            raise InvalidInputException("invalid_dataframe")

        return output


class TensorInput(ModelInputSerializer):
    def define(self, shape=None, dtype=float):
        self.shape = shape
        self.dtype = dtype

    def serialize(self, input):
        try:
            torch = import_module("torch")
        except ModuleNotFoundError:
            raise RuntimeError("torch_required_for_tensor_inputs")

        output = torch.tensor(input, dtype=self.dtype)

        errors = []
        if self.shape is not None and output.shape != self.shape:
            errors = ["invalid_shape"]

        if errors:
            raise InvalidInputException("invalid_tensor_shape", errors=errors)

        return output


class ModelInputs:
    Float = FloatInput
    String = StringInput
    NumpyArray = NumpyInput
    Dataframe = DataframeInput
    Tensor = TensorInput

    @staticmethod
    def validate_input_keys(inputs, expected_inputs):
        errors = []

        if not isinstance(inputs, dict):
            raise InvalidInputException("invalid_input_format")

        for key in expected_inputs.keys():
            if key not in inputs.keys():
                errors.append(f"missing input: {key}")

        if errors:
            raise InvalidInputException("invalid_input_format", errors=errors)

        return True



import yaml

from slai.clients.cli import get_cli_client
from slai.modules.parameters import from_config

from pathlib import Path
from importlib import import_module


def get_project_client(*, project_name):
    import_path = from_config(
        "PROJECT_CLIENT",
        "slai.clients.project.ProjectClient",
    )
    class_ = import_path.split(".")[-1]
    path = ".".join(import_path.split(".")[:-1])

    return getattr(import_module(path), class_)(project_name=project_name)


class ProjectClient:
    def __init__(self, *, project_name=None):
        self.project_name = project_name
        self.cli_client = get_cli_client()

    def list_models(self):
        """List models in a slai project."""

        project = self.get_project()
        return project["models"]

    def _load_config_from_disk(self):
        cwd = Path.cwd()
        config = None

        local_config_path = f"{cwd}/.slai/config.yml"

        with open(local_config_path, "r") as f_in:
            try:
                config = yaml.safe_load(f_in)
            except yaml.YAMLError:
                pass

        return config

    def get_project_name(self):
        """Retrieve project name."""
        if self.project_name is None:
            config = self._load_config_from_disk()
            return config["project_name"]

        return self.project_name

    def get_project(self):
        """Retrieve project configuration values."""
        project = self.cli_client.retrieve_project(project_name=self.get_project_name())

        return project



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['slai', 'slai.clients', 'slai.loaders', 'slai.modules']

package_data = \
{'': ['*'], 'slai': ['templates/*']}

install_requires = \
['GitPython>=3.1.11,<4.0.0',
 'Jinja2>=2.11.2,<3.0.0',
 'PyDrive>=1.3.1,<2.0.0',
 'boto3>=1.16.44,<2.0.0',
 'celery==5.0.2',
 'click>=7.1.2,<8.0.0',
 'python-dotenv>=0.15.0,<0.16.0',
 'redis>=3.5.3,<4.0.0',
 'shtab>=1.3.4,<2.0.0',
 'tqdm>=4.61.1,<5.0.0',
 'xxhash>=2.0.2,<3.0.0']

setup_kwargs = {
    'name': 'slai',
    'version': '0.1.44',
    'description': '',
    'long_description': None,
    'author': 'slai',
    'author_email': 'luke@slai.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.1,<4.0.0',
}


setup(**setup_kwargs)
